ALLOWED_BANDS = ["g", "r"]

FILTERS = ["g", "r"]
FILTER2IDX = {"g": 1, "r": 2}

BIN_IDX_TO_FREQ_COL = {0: "frequency_20", 1: "frequency_10", 2: "frequency_5"}
BIN_IDX_TO_FAP_COL = {0: "FAP_20", 1: "FAP_10", 2: "FAP_5"}
